"""Shared viewer core and optional desktop frontend."""

from .document import (
    StateEntry,
    ViewerDocument,
    load_npz_document,
    load_sad_document,
    load_viewer_document,
)
from .session import (
    DEFAULT_NUMERIC_STEP,
    PROJECTION_LABELS,
    ImageContainerHost,
    ViewerFrame,
    ViewerHost,
    ViewerSession,
    ViewerSessionResidencySnapshot,
    _coerce_director_field_state,
    _decorate_property_snapshot,
    _effective_numeric_step,
    _property_snapshot,
    _quantize_numeric_value,
    adapt_viewer_host,
    apply_camera_command,
    camera_state_signature,
    format_camera_summary,
    infer_cubic_system_from_state,
    restore_camera,
    serialize_camera,
)
from .residency import SceneResidencyManager, system_cache_key

__all__ = [
    "DEFAULT_NUMERIC_STEP",
    "PROJECTION_LABELS",
    "ImageContainerHost",
    "SceneResidencyManager",
    "StateEntry",
    "ViewerDocument",
    "ViewerFrame",
    "ViewerHost",
    "ViewerSession",
    "ViewerSessionResidencySnapshot",
    "_coerce_director_field_state",
    "_decorate_property_snapshot",
    "_effective_numeric_step",
    "_property_snapshot",
    "_quantize_numeric_value",
    "adapt_viewer_host",
    "apply_camera_command",
    "camera_state_signature",
    "format_camera_summary",
    "infer_cubic_system_from_state",
    "load_npz_document",
    "load_sad_document",
    "load_viewer_document",
    "restore_camera",
    "serialize_camera",
    "system_cache_key",
]
